/// lib.rs

use pyo3::prelude::*;
use pyo3::exceptions::PyRuntimeError;
use postgres::{Client, NoTls};
use serde::{Serialize, Deserialize};  /// Import derive macros from serde_derive

/// Define our User model, which will be exposed to Python.
#[pyclass]
#[derive(Serialize, Deserialize, Debug)]
struct User {
    #[pyo3(get, set)]
    id: Option<i32>,
    #[pyo3(get, set)]
    name: String,
    #[pyo3(get, set)]
    email: String,
}
/// Group Model
#[pyclass]
#[derive(Serialize, Deserialize, Debug)]
struct Group {
    #[pyo3(get, set)]
    id: Option<i32>,
    #[pyo3(get, set)]
    name: String,
}

/// A helper function to convert postgres::Error into a Python RuntimeError.
fn pg_err(e: postgres::Error) -> PyErr {
    PyRuntimeError::new_err(e.to_string())
}

/// Initialize the database by creating the users table if it does not exist.
#[pyfunction]
fn init_db(db_url: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.batch_execute(
        "
        CREATE TABLE IF NOT EXISTS users (
            id SERIAL PRIMARY KEY,
            name VARCHAR NOT NULL,
            email VARCHAR NOT NULL
        );
        CREATE TABLE IF NOT EXISTS groups (
            id SERIAL PRIMARY KEY,
            name VARCHAR NOT NULL
        );
        CREATE TABLE IF NOT EXISTS group_members (
            group_id INTEGER REFERENCES groups(id),
            user_id INTEGER REFERENCES users(id),
            PRIMARY KEY (group_id, user_id)
        );
        "
    ).map_err(pg_err)?;
    Ok(())
}

/// Create a new user by inserting into the database.
#[pyfunction]
fn create_user(db_url: &str, name: &str, email: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO users (name, email) VALUES ($1, $2)",
        &[&name, &email]
    ).map_err(pg_err)?;
    Ok(())
}

/// Retrieve a user by ID. Returns None if the user is not found.
#[pyfunction]
fn get_user(db_url: &str, user_id: i32) -> PyResult<Option<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name, email FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let user = User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
        };
        Ok(Some(user))
    } else {
        Ok(None)
    }
}

/// Retrieve all users from the database.
#[pyfunction]
fn get_all_users(db_url: &str) -> PyResult<Vec<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name, email FROM users", &[])
        .map_err(pg_err)?;
    
    let mut users = Vec::new();
    for row in rows {
        users.push(User {
            id: row.get(0),
            name: row.get(1),
            email: row.get(2),
        });
    }
    Ok(users)
}

/// Update an existing user with a new name and email.
#[pyfunction]
fn update_user(db_url: &str, user_id: i32, new_name: &str, new_email: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let updated = client.execute(
        "UPDATE users SET name = $1, email = $2 WHERE id = $3",
        &[&new_name, &new_email, &user_id]
    ).map_err(pg_err)?;
    if updated == 0 {
        Err(PyRuntimeError::new_err("User not found"))
    } else {
        Ok(())
    }
}

/// Delete a user by ID. Returns true if a user was deleted.
#[pyfunction]
fn delete_user(db_url: &str, user_id: i32) -> PyResult<bool> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let deleted = client.execute(
        "DELETE FROM users WHERE id = $1",
        &[&user_id]
    ).map_err(pg_err)?;
    Ok(deleted > 0)
}

/// Create a group
#[pyfunction]
fn create_group(db_url: &str, name: &str) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO groups (name) VALUES ($1)",
        &[&name]
    ).map_err(pg_err)?;
    Ok(())
}

/// Retrieve a group by ID.
#[pyfunction]
fn get_group(db_url: &str, group_id: i32) -> PyResult<Option<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let row_opt = client.query_opt(
        "SELECT id, name FROM groups WHERE id = $1",
        &[&group_id]
    ).map_err(pg_err)?;
    
    if let Some(row) = row_opt {
        let group = Group {
            id: row.get(0),
            name: row.get(1),
        };
        Ok(Some(group))
    } else {
        Ok(None)
    }
}

/// Retrieve all groups.
#[pyfunction]
fn get_all_groups(db_url: &str) -> PyResult<Vec<Group>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query("SELECT id, name FROM groups", &[])
        .map_err(pg_err)?;
    
    let groups = rows.into_iter().map(|row| Group {
        id: row.get(0),
        name: row.get(1),
    }).collect();
    
    Ok(groups)
}

/// Add a user to a group.
#[pyfunction]
fn add_user_to_group(db_url: &str, group_id: i32, user_id: i32) -> PyResult<()> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    client.execute(
        "INSERT INTO group_members (group_id, user_id) VALUES ($1, $2)",
        &[&group_id, &user_id]
    ).map_err(pg_err)?;
    Ok(())
}

/// Get all members of a group.
#[pyfunction]
fn get_group_members(db_url: &str, group_id: i32) -> PyResult<Vec<User>> {
    let mut client = Client::connect(db_url, NoTls).map_err(pg_err)?;
    let rows = client.query(
        "SELECT u.id, u.name, u.email
         FROM users u
         JOIN group_members gm ON u.id = gm.user_id
         WHERE gm.group_id = $1",
         &[&group_id]
    ).map_err(pg_err)?;
    
    let users = rows.into_iter().map(|row| User {
        id: row.get(0),
        name: row.get(1),
        email: row.get(2),
    }).collect();
    
    Ok(users)
}


/// This is the Python module initializer. The name "rust_crud_api" here should match the name used in your Cargo.toml.
#[pymodule]
fn rust_crud_api(_py: Python, m: &PyModule) -> PyResult<()> {
    m.add_class::<User>()?;
    m.add_class::<Group>()?;
    m.add_function(wrap_pyfunction!(init_db, m)?)?;
    m.add_function(wrap_pyfunction!(create_user, m)?)?;
    m.add_function(wrap_pyfunction!(get_user, m)?)?;
    m.add_function(wrap_pyfunction!(get_all_users, m)?)?;
    m.add_function(wrap_pyfunction!(update_user, m)?)?;
    m.add_function(wrap_pyfunction!(delete_user, m)?)?;
    m.add_function(wrap_pyfunction!(create_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_all_groups, m)?)?;
    m.add_function(wrap_pyfunction!(add_user_to_group, m)?)?;
    m.add_function(wrap_pyfunction!(get_group_members, m)?)?;
    Ok(())
}

